/*  Audacious-DUMB Plug-in to enable audacious to use DUMB for IT/XM/S3M/MOD files
 *  Copyright (C) 2006-2007 Christian Birchinger
 *  Audacious port based on previous plugins for XMMS and BMP
 *  Copyright (C) 2002-2003  Ben Davis
 *  Incorporates code from the wav plug-in,
 *  Copyright (C) 1998-2000  Peter Alm, Mikael Alm, Olle Hallnas, Thomas Nilsson and 4Front Technologies
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "audacious-dumb.h"

#include <audacious/configdb.h>
#include <audacious/vfs.h>
#include <audacious/output.h>

#include <gtk/gtk.h>
#include "interface.h"
#include "support.h"

DuhFile *duh_file = NULL;
static GThread *decode_thread;
static gboolean disable_amiga_mods = FALSE;
static const gchar duh_custom_title_format[] = "%f%{t: - %t%}";
static gint output_frequency;
static const gint output_frequencies[]={4000, 8000, 11025, 22050, 44100, 48000, 64000, 96000};
static GtkWidget *prefs;

gchar *fmts[] = { "mod", "s3m", "it", "xm", "duh", NULL };

InputPlugin duh_ip =
{
	NULL,
	NULL,
	"DUMB Plugin " VERSION,
	duh_init,
	duh_about,
	duh_configure,
	is_our_file,
	NULL,
	play_file,
	stop,
	duh_pause,
	seek,
	NULL,
	NULL,
	NULL,
	NULL,
	dumb_exit,
	NULL,
	NULL,
	NULL,
	NULL,
	get_song_info,
	file_info_box,
	NULL,
	get_tuple_info,
	NULL,
	NULL,
	is_our_file_from_vfs,
	fmts,
};

InputPlugin *get_iplugin_info(void)
{
	duh_ip.description = VERSIONSTRING;
	return &duh_ip;
}

static void duh_init(void)
{
	ConfigDb *db;
	db = bmp_cfg_db_open();

	bmp_cfg_db_get_bool(db, "DUMB", "disable_amiga_mods", &disable_amiga_mods);
	if (!disable_amiga_mods) /* force db creation since defined false equals none-existant */
		bmp_cfg_db_set_bool(db, "DUMB", "disable_amiga_mods", FALSE);

	bmp_cfg_db_get_int(db, "DUMB", "output_frequency", &output_frequency);	
	if (output_frequency < 4000 || output_frequency > 96000) {
		output_frequency = 44100;
		bmp_cfg_db_set_int(db, "DUMB", "output_frequency", output_frequency);
	}
		
	bmp_cfg_db_close(db);

	dumb_register_stdfiles();
	dumb_it_max_to_mix = 256;
}

static int is_our_file(char *filename)
{
	const gchar *ext;

	ext = strrchr(filename, '.');
	if (ext) {
		if (!strcasecmp(ext, ".duh")) return TRUE;
		if (!strcasecmp(ext, ".it")) return TRUE;
		if (!strcasecmp(ext, ".xm")) return TRUE;
		if (!strcasecmp(ext, ".s3m")) return TRUE;
		if (!disable_amiga_mods) {
			if (!strcasecmp(ext, ".mod")) return TRUE;
		}
	}
	return FALSE;
}

static int is_our_file_from_vfs(gchar *filename, VFSFile *file)
{
	gchar magic[4];
	const gchar *ext;

	vfs_fread(magic, 1, 4, file);
	if (!memcmp(magic, XM_MAGIC, 4))
                return TRUE;
	if (!memcmp(magic, IT_MAGIC, 4))
		return TRUE;

	vfs_fseek(file, 44, SEEK_SET);
	vfs_fread(magic, 1, 4, file);
	if (!memcmp(magic, S3M_MAGIC, 4))
		return TRUE;

	vfs_fseek(file, 1080, SEEK_SET);
	vfs_fread(magic, 1, 4, file);   
	if (!memcmp(magic, MOD_MAGIC_FASTTRACKER6, 4))
		return TRUE;
	if (!memcmp(magic, MOD_MAGIC_FASTTRACKER8, 4))
		return TRUE;

	if (!disable_amiga_mods) {
		if (!memcmp(magic, MOD_MAGIC_PROTRACKER4, 4))
			return TRUE;
		if (!memcmp(magic, MOD_MAGIC_PROTRACKER4X, 4))
			return TRUE;
		if (!memcmp(magic, MOD_MAGIC_NOISETRACKER, 4))
			return TRUE;
		if (!memcmp(magic, MOD_MAGIC_STARTRACKER4, 4))
			return TRUE;
		if (!memcmp(magic, MOD_MAGIC_STARTRACKER8, 4))
			return TRUE;
		if (!memcmp(magic, MOD_MAGIC_STARTRACKER4X, 4))
			return TRUE;
		if (!memcmp(magic, MOD_MAGIC_STARTRACKER8X, 4))
			return TRUE;
		if (!memcmp(magic, MOD_MAGIC_FASTTRACKER4, 4))
			return TRUE;
	}

	ext = strrchr(filename, '.');
	if (ext) {
		if (!strcasecmp(ext, ".duh")) return TRUE;
		if (!strcasecmp(ext, ".it")) return TRUE;
		if (!strcasecmp(ext, ".xm")) return TRUE;
		if (!strcasecmp(ext, ".s3m")) return TRUE;
		if (!disable_amiga_mods) {
			if (!strcasecmp(ext, ".mod")) return TRUE;
		}
	}

	return FALSE;
}

static gint duh_universal_load_vfs(DUH **duh, const char *filename, const short testmask)
{
	DUMBFILE *filedumb;

	gint ext_index = -1;
	VFSFile *file;
	glong filesize;
	char *filemap;

	if (!(file = vfs_fopen(filename, "rb")))
		return -1;

	vfs_fseek(file, 0, SEEK_END);
	filesize = vfs_ftell(file);
	vfs_fseek(file, 0, SEEK_SET);

	if (!(filemap = malloc(filesize)))
		return -1;

	vfs_fread (filemap, 1, filesize, file);
	vfs_fclose(file);

	/*
	   The close and re-opening after reach read attempt is needed.

	   Quote from the documentation:

	   WARNING: The behaviour of these functions are undefined if you pass a
	   DUMBFILE from which data have already been read; they are likely
	   not to work. This oversight will be fixed in future releases.
	*/

	int count, total_count = sizeof(uniread) / sizeof(uniread_struct_t);

	for (count = 0; count < total_count; count++) {
		if (uniread[count].testmask & testmask) {
			filedumb = dumbfile_open_memory(filemap, filesize);
			*duh = uniread[count].read(filedumb);
			dumbfile_close(filedumb);
			if (*duh) {
				ext_index = count;
				break;
			}
		}
	}

	free(filemap);
	filemap = NULL;

	return ext_index;
}

static TitleInput *get_tuple_info_from_duh(DUH *duh, char *filename)
{
	const gchar *mod_title_raw;
	gchar *mod_title;
	TitleInput *tuple = bmp_title_input_new();
	
	tuple->file_name = g_path_get_basename(filename);
	tuple->file_path = g_path_get_dirname(filename);
	tuple->formatter = g_strdup(duh_custom_title_format);

	if (duh) {
		tuple->length = (int)(duh_get_length(duh) * 1000LL >> 16);
		
		mod_title_raw = duh_get_tag(duh, "TITLE");
		if (mod_title_raw) {
			/* DOS (CP850) all my XM IT and S3M use this (MODs should be 7bit ASCII) */ 
			mod_title = g_convert(mod_title_raw, -1, "UTF-8", "CP850", NULL, NULL, NULL);
			/* Raw copy but GTK2 doesn't like none-UTF8 stuff*/ 
			/* mod_title = g_strdup(mod_title_raw); */

			g_strstrip(mod_title);

			if (mod_title[0])
				tuple->track_name = g_strdup(mod_title);

			g_free(mod_title);
			mod_title = NULL;
		}
		mod_title_raw = NULL;
	}
	return tuple;
}

static TitleInput *get_tuple_info(char *filename)
{
	TitleInput *input = NULL;
	DUH *duh;
	gint ext_index;

	ext_index = duh_universal_load_vfs(&duh, filename, UNIREAD_ALL);
	
	input = get_tuple_info_from_duh(duh, filename);

	if (duh) {
		input->file_ext = g_strdup( uniread[ext_index].ext );
		unload_duh(duh);
	}

	return input;
}

static void get_song_info(char *filename, char **title, int *length)
{
	TitleInput *tuple = get_tuple_info(filename);
	
	*length = tuple->length;
	*title = xmms_get_titlestring(duh_custom_title_format, tuple);
	
	bmp_title_input_free(tuple);
}

static gchar *get_title(DUH *duh, char *filename)
{
        TitleInput *tuple = NULL;
	gchar *title;
	
	tuple = get_tuple_info_from_duh(duh, filename);
	title = xmms_get_titlestring(duh_custom_title_format, tuple);

	bmp_title_input_free(tuple);

	return title;
}

static void install_callbacks(DUH_SIGRENDERER *sr)
{
	DUMB_IT_SIGRENDERER *itsr = duh_get_it_sigrenderer(sr);
	dumb_it_set_loop_callback(itsr, &dumb_it_callback_terminate, NULL);
	dumb_it_set_xm_speed_zero_callback(itsr, &dumb_it_callback_terminate, NULL);
}

static void *play_loop(void *arg)
{
	char data[2048 * 2];
	int blk_size, rate;
	unsigned int bytes;
	int actual_read;
	
	InputPlayback *playback = arg;

	blk_size = 512 * (duh_file->bits_per_sample / 8) * duh_file->channels;
	rate = duh_file->samples_per_sec * duh_file->channels * (duh_file->bits_per_sample / 8);
	while (playback->playing)
	{
		if (!playback->eof)
		{
			bytes = blk_size;
			if (duh_file->length - duh_file->position < bytes)
				bytes = duh_file->length - duh_file->position;
			if (bytes > 0)
			{
				actual_read = duh_render(duh_file->file, 16, 0, 1.0f, 65536.0f / duh_file->samples_per_sec, bytes / ((duh_file->bits_per_sample / 8) * duh_file->channels), data);
				actual_read *= (duh_file->bits_per_sample / 8) * duh_file->channels;

				if (actual_read == 0)
				{
					/*  Should we do this or should we pad with silence? */
					playback->eof = TRUE;
					playback->output->buffer_free();
					playback->output->buffer_free();
				}
				else
				{
					if (playback->playing && duh_file->seek_to == -1)
						produce_audio(playback->output->written_time(), (duh_file->bits_per_sample == 16) ? FMT_S16_LE : FMT_U8,
						duh_file->channels, bytes, data, &playback->playing);

					duh_file->position += actual_read;
				}
			}
			else
			{
				playback->eof = TRUE;
				playback->output->buffer_free();
				playback->output->buffer_free();
				g_usleep(10000);
			}
		}
		else
			g_usleep(10000);
			
		if (duh_file->seek_to != -1)
		{
			duh_file->position = duh_file->seek_to * rate;
			duh_end_sigrenderer(duh_file->file);
			duh_file->file = duh_start_sigrenderer(duh_file->duh, 0, duh_file->channels, duh_file->seek_to << 16);
			install_callbacks(duh_file->file);
			playback->output->flush(duh_file->seek_to * 1000);
			duh_file->seek_to = -1;
		}

	}
	duh_end_sigrenderer(duh_file->file);
	unload_duh(duh_file->duh);
	g_thread_exit(NULL);
	
	return NULL;
}

static void play_file(InputPlayback *playback)
{
	char *name;
	int rate;
	char *filename = playback->filename;

	playback->error = FALSE;

	duh_file = g_malloc(sizeof (DuhFile));
	memset(duh_file, 0, sizeof (DuhFile));

	duh_universal_load_vfs(&duh_file->duh, filename, UNIREAD_ALL);
	if (!duh_file->duh) {
		printf ("DUMB error: unable to play %s\n", filename);
		g_free(duh_file);
		duh_file = NULL;
		return;
	}

	if ((duh_file->file = duh_start_sigrenderer(duh_file->duh, 0, 2, 0)))
	{
		install_callbacks(duh_file->file);
		duh_file->channels = 2;
		duh_file->samples_per_sec = output_frequency;
		duh_file->bits_per_sample = 16;
		duh_file->length = (long long)duh_get_length(duh_file->duh) * (output_frequency * 2 * 2) >> 16;
		duh_file->position = 0;
		playback->playing = 1;

		if (playback->output->open_audio((duh_file->bits_per_sample == 16) ? FMT_S16_LE : FMT_U8, duh_file->samples_per_sec, duh_file->channels) == 0)
		{
			playback->error = TRUE;
			duh_end_sigrenderer(duh_file->file);
			unload_duh(duh_file->duh);
			g_free(duh_file);
			duh_file = NULL;
			return;
		}
		name = get_title(duh_file->duh, filename);
		rate = duh_file->samples_per_sec * duh_file->channels * (duh_file->bits_per_sample / 8);
		duh_ip.set_info(name, 1000LL * duh_file->length / rate, 8 * rate, duh_file->samples_per_sec, duh_file->channels);
		g_free(name);
		duh_file->seek_to = -1;
		decode_thread = g_thread_create(play_loop, playback, TRUE, NULL);
	}
}

static void stop(InputPlayback *playback)
{
	if (duh_file && playback->playing)
	{
		playback->playing = 0;
		g_thread_join(decode_thread);
		playback->output->close_audio();
		g_free(duh_file);
		duh_file = NULL;
	}
}

static void duh_pause(InputPlayback *playback, short p)
{
	playback->output->pause(p);
}

static void seek(InputPlayback *playback, int time)
{
	duh_file->seek_to = time;

	playback->eof = FALSE;

	while (duh_file->seek_to != -1)
		g_usleep(10000);
}

static void file_info_box(char *filename)
{
	GtkWidget *songinfo;
	GtkStyle *style = NULL;
	GtkTextBuffer *TextBuffer;
	PangoFontDescription *font_desc = NULL;

	DUMB_IT_SIGDATA *sd;
	DUH *duh;

	if ( duh_universal_load_vfs(&duh, filename, UNIREAD_MOD) == -1 )
		return;
	
	sd = duh_get_it_sigdata(duh);

	songinfo = create_songinfo ();
	if (!songinfo) return;

	font_desc = pango_font_description_from_string ("monospace 9");

	if (font_desc) {
		style = gtk_style_copy(gtk_widget_get_style(songinfo));
		style->font_desc = font_desc;
	}
	
	{
		int n = dumb_it_sd_get_n_instruments(sd);
		if (n) {
			GtkCList *l = GTK_CLIST(lookup_widget(songinfo, "songinfo_instruments_clist"));
			int i;
			if (style) gtk_widget_set_style(GTK_WIDGET(l), style);
			gtk_clist_column_titles_passive(l);
			gtk_clist_set_column_justification(l, 0, GTK_JUSTIFY_RIGHT);
			for (i = 0; i < n; i++) {
				char str[3];
				const gchar *text[] = {
					str,
					(gchar*) dumb_it_sd_get_instrument_name(sd, i),
					(gchar*) dumb_it_sd_get_instrument_filename(sd, i)
				};
				sprintf(str, "%d", i+1);
				gtk_clist_insert(l, i, (gchar **)text);
			}
		} else
			gtk_notebook_remove_page(GTK_NOTEBOOK(lookup_widget(songinfo, "songinfo_notebook")), 2);
	}

	{
		GtkCList *l = GTK_CLIST(lookup_widget(songinfo, "songinfo_samples_clist"));
		int i, n = dumb_it_sd_get_n_samples(sd);
		if (style) gtk_widget_set_style(GTK_WIDGET(l), style);
		gtk_clist_column_titles_passive(l);
		gtk_clist_set_column_justification(l, 0, GTK_JUSTIFY_RIGHT);
		for (i = 0; i < n; i++) {
			char str[3];
			const gchar *text[] = {
				str,
				(gchar*) dumb_it_sd_get_sample_name(sd, i),
				(gchar*) dumb_it_sd_get_sample_filename(sd, i)
			};
			sprintf(str, "%d", i+1);
			gtk_clist_insert(l, i, (gchar **)text);
		}
	}

	{
		GtkTextView *l = GTK_TEXT_VIEW(lookup_widget(songinfo, "songinfo_message_text"));
		const char *message = (char*) dumb_it_sd_get_song_message(sd);
		if (message) {
			char *m = strdup(message);
			int i;
			for (i = 0; m[i]; i++)
				if (m[i] == '\r') m[i] = '\n';

			TextBuffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (lookup_widget(songinfo, "songinfo_message_text")));
			if (style) gtk_widget_set_style(GTK_WIDGET(l), style);
			gtk_text_buffer_set_text(TextBuffer, (gchar*) m, -1);

			free(m);
		} else
			gtk_notebook_remove_page(GTK_NOTEBOOK(lookup_widget(songinfo, "songinfo_notebook")), 0);
	}

	/* I hope that's ok here. Previously there was no unload_duh at all (leak?) */
	if (duh)
		unload_duh(duh);

	gtk_widget_show (songinfo);
}

static void duh_configure(void) {
	if (prefs) return;
	prefs = create_prefs();
	if (!prefs) return;

	gint output_frequency_idx = 4; /* 44100 default/fallback */
	int i;
	gchar freqstring[6];

	for (i = 0; i < (int) (sizeof(output_frequencies) / sizeof(output_frequencies[0])); i++) {
		snprintf(freqstring, sizeof(freqstring), "%d", output_frequencies[i]);
		gtk_combo_box_append_text (GTK_COMBO_BOX(lookup_widget(prefs, "prefs_freq_combobox")), freqstring);
		if (output_frequencies[i] == output_frequency)
			output_frequency_idx = i;
	}
	
 	gtk_combo_box_set_active(GTK_COMBO_BOX(lookup_widget(prefs, "prefs_freq_combobox")), output_frequency_idx);

	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(lookup_widget(prefs, "prefs_checkbox_disableamiga")), disable_amiga_mods);

	gtk_signal_connect(GTK_OBJECT(prefs), "destroy",         
			   GTK_SIGNAL_FUNC(gtk_widget_destroyed), &prefs);

	gtk_signal_connect_object(GTK_OBJECT(lookup_widget(prefs, "prefs_button_cancel")), "clicked",
				  GTK_SIGNAL_FUNC(gtk_widget_destroy), 
				  GTK_OBJECT(prefs));

	gtk_signal_connect(GTK_OBJECT(lookup_widget(prefs, "prefs_button_ok")), "clicked",
				  GTK_SIGNAL_FUNC(duh_configure_ok_cb), lookup_widget(prefs, "prefs_checkbox_disableamiga"));

	gtk_widget_show(prefs);
}

static void duh_configure_ok_cb(GtkButton *w, gpointer data) {
	ConfigDb *db;
	gint output_frequency_idx;

	db = bmp_cfg_db_open();

	disable_amiga_mods = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(data));
	bmp_cfg_db_set_bool(db, "DUMB", "disable_amiga_mods", disable_amiga_mods);

	output_frequency_idx = gtk_combo_box_get_active(GTK_COMBO_BOX(lookup_widget(prefs, "prefs_freq_combobox")));
	output_frequency = output_frequencies[output_frequency_idx];
	bmp_cfg_db_set_int(db, "DUMB", "output_frequency", output_frequency);	

	bmp_cfg_db_close(db);
	gtk_widget_destroy(prefs);
}

static void duh_about(void) {
	static GtkWidget *about;

	if (about) return;

	about = xmms_show_message("About DUMB Plugin",
				  "DUMB Input Plugin " VERSION "\n"
				  "by Christian Birchinger <joker@netswarm.net>\n"
				  "\n"
				  "Based on the original XMMS plugin by Ben Davis\n"
				  "and the BMP port by Michael Doering\n"
				  "\n"
				  "Built with DUMB version " DUMB_VERSION_STR,
				  "Ok", FALSE, NULL, NULL);

	g_signal_connect(G_OBJECT(about), "destroy",
			 G_CALLBACK(gtk_widget_destroyed), &about);
}
